import os
import sys
from transformation import *
from structure import *
from parserF import read_formula
from nnf import toNNF
import time


# On a ici la procedure principale de model-checking.
# On lit notre structure et notre formule a partir d'un fichier,
# puis on les transforme en une formule QBF. On ecrit cette formule
# dans un fichier temporaire, puis on appelle Z3 dessus.

# Dans notre programme on ne soumet a Z3 qu'un axiome.
# J'ai toutefois ecrit une fonction plus generale qui permet
# de soumettre a Z3 une theorie de plusieurs axiomes
# (ce qui revient a prendre la conjonction de tous les axiomes).
# Je permet aussi d'utiliser des variables libres (ce qui revient,
# pour Z3, a mettre un quantificateur existentiel derriere).
def write_Z3(theory, pathname) :
    output=open(pathname, 'w')

    # On construit l'ensemble de toutes les variables libres,
    # qu'il faudra declarer a Z3 comme des constantes.
    varsAux=[]
    def aux(f):
        if isinstance(f, AP) and (not isinstance(f, QUANT)) :
            varsAux.append(f.toString())
        for f2 in f.fils :
            aux(f2)
    for axiom in theory :
        aux(axiom)
    vars2=frozenset(varsAux)

    # Declaration des constantes.
    for v in vars2 :
        output.write("(declare-const "+v+" Bool)\n")

    # Declaration des axiomes.
    for axiom in theory :
        output.write("(assert ")
        axiom.write(output)
        output.write(")\n")

    output.write("(check-sat)\n")
    output.close()

def give_Z3(pathname) :
	cmd=cz3 + " " + pathname
	if os.name != 'nt' :
	    cmd="time " + cmd
	os.system(cmd)

# out est le nom du fichier dans lequel on ecrira la formule QBF.
# mode est le numero de la methode a utiliser.
def model_checking(krip, formule, out, mode) :
    if mode == 5 : # procedure speciale pour Prenex NF
        model_checking_PNF(krip, formule, out)
        return
    
    start=time.time()
    z3Form=toQBF(krip, krip.start, formule, mode)
    end=time.time()
    print("Time needed for the construction f the QBF formula : "+str(end-start))
    
    print("Size of the formula           : "+str(findLength(z3Form)))
#    print("Nb d'alternations dans la formule QBF: ",z3Form.nbalt())

    start=time.time()
    #print("Formule avant le write:")
    #print(z3Form.toString())
    write_Z3([z3Form], out)
    end=time.time()
    print("Time needed to write the formula in the file          : "+str(end-start))

    start=time.time()
    give_Z3(out)
    end=time.time()
    print("Time needed for Z3                   : "+str(end-start))

if __name__ == "__main__" :
    model_checking(Krip.from_file(sys.argv[1]), read_formula(sys.argv[2]), sys.argv[3], int(sys.argv[4]))


    
####################################################################################

def MethDecode(methode) :
    if methode=='UU' :
        print("Method : unfolding untils")
        return 3
    elif methode=='FP' :
        print("Method : fixed point charac.")
        return 4
    elif methode=='FFP' :
        print("Method : flattening + fixed point charac. ")
        return 7
    elif methode=='FBV' :
        print("Method : flattening + fixed point charac and bitvectors.")
        return 5
    else :
        print("Unknown method:",methode)
        return 3

def verif(nameK,nameF,pref="TEST/",methode='UU',buntil=0) :
    print("-> QCTL model-checking (V1.0 - June 2019)")
    nmeth=MethDecode(methode)
    F=read_formula(pref+nameF)
    if nmeth>5 :
        F = F.FNN(False)
        F=F.SimplifB()
        F=F.MaPl({},{},-1,False)
        print("Formula : ",F.toString())
        nmeth=nmeth-3 # 6 -> methode 3, 7 -> methode 4
    elif nmeth==5 and buntil>0 :
        print("Bounded untils :",str(buntil))
    S=Krip.from_file(pref+nameK,buntil)
    model_checking(S,F,pref+nameK+nameF+'.z3',nmeth)
    
####################################################################################

