from formule import *

# Le but de ce module est de pouvoir mettre une formule sous forme normale negative.

# L'idee est d'"enfoncer" les negations d'une formule le plus profondement possible dans son AST,
# jusqu'a ce que celles-ci setrouvent juste derriere les feuilles.

# negate sert a les "enfoncer" d'au moins un niveau.

# toNNF est une fonction recursive qui se sert de negate
# pour obtenir la forme normale negative de notre formule.

def negate(formule) :
    if isinstance(formule, AP) :
        return NOT(formule)

    elif isinstance(formule, NOT) :
        return formule.fils[0]

    elif isinstance(formule, AND) :
        return OR(negate(formule.fils[0]), negate(formule.fils[1]))

    elif isinstance(formule, OR) :
        return AND(negate(formule.fils[0]), negate(formule.fils[1]))

    elif isinstance(formule, EX) :
        return AX(negate(formule.fils[0]))

    elif isinstance(formule, AX) :
        return EX(negate(formule.fils[0]))

    elif isinstance(formule, EU) :
        return AW(negate(formule.fils[1]), negate(formule.fils[0]))

    elif isinstance(formule, AU) :
        return EW(negate(formule.fils[1]), negate(formule.fils[0]))

    elif isinstance(formule, EW) :
        return AU(negate(formule.fils[1]), negate(formule.fils[0]))

    elif isinstance(formule, AW) :
        return EU(negate(formule.fils[1]), negate(formule.fils[0]))

    elif isinstance(formule, EF) :
        return AG(negate(formule.fils[0]))

    elif isinstance(formule, EG) :
        return AF(negate(formule.fils[0]))

    elif isinstance(formule, AF) :
        return EG(negate(formule.fils[0]))

    elif isinstance(formule, AG) :
        return EF(negate(formule.fils[0]))

    elif isinstance(formule, TOP) :
        return BOTTOM()

    elif isinstance(formule, BOTTOM) :
        return TOP()

    elif isinstance(formule, FORALL) :
        return EXISTS(formule.var, negate(formule.fils[0]))

    elif isinstance(formule, EXISTS) :
        return FORALL(formule.var, negate(formule.fils[0]))

    elif isinstance(formule, FORALL1) :
        return EXISTS1(formule.var, negate(formule.fils[0]))

    elif isinstance(formule, EXISTS1) :
        return FORALL1(formule.var, negate(formule.fils[0]))
    
    elif isinstance(formule, SEQAND) :
        newList=[]
        for fils in formule.fils :
            newList.append(negate(fils))
        return SEQOR(newList)

    elif isinstance(formule, SEQOR) :
        newList=[]
        for fils in formule.fils :
            newList.append(negate(fils))
        return SEQAND(newList)

    elif isinstance(formule, IMPL) :
        return AND(formule.fils[0], negate(formule.fils[1]))

    # On ne fait rien pour l'egalite.
    # Le but de la mise sous nnf est de simplifier la formule QBF qu'on soumet a z3.
    # La nnf de la negation d'une equivalence est une formule plus complexe que la negation.
    # Par exemple, la nnf (!a & b) or (a & !b) est plus complexe que !(a == b).

def toNNF(formule) :
    ans=formule
    while isinstance(ans, NOT) and not isinstance(ans.fils[0], EQ) :
        if isinstance(ans.fils[0], AP) :
            return ans
        ans=negate(ans.fils[0])
    for i in range(len(ans.fils)) :
        ans.fils[i]=toNNF(ans.fils[i])
    return ans
